

import java.awt.*;

class Road {
  // Read Only variables for display:

  // length of the road
  public int LENGTH = 400;
  // Maximum speed
  public int MAXSPEED;
  // the road as an array of speeds
  public int[] speed;
  // local density of vehicles on the road [veh/site]
  public double dens;
  // local mean velocity [sites/timestep]
  public double v;

  // global frequency of gaps
  public double[] gapfreq;
  // global frequency of velocities
  public double[] vfreq;

  // dummy for infinity
  private final double DBL_INF = 99999.99;

  // number of vehicles on the road
  private int cars;
  // 1/number of vehicles
  private double invcars;
  // cell length
  private double deltaX;
  private int[] ve;

  // my private colors
  private final Color streetcolor = Color.gray;
  private final Color background = Color.white;
  private final Color foreground = Color.black;
  private final Color linecolor = Color.blue;
  private final Color vcolor[];

  //private Road() {}

  public Road (double density, double deltaX) {
    this.deltaX = deltaX;
    this.MAXSPEED = (int)(135.0/(deltaX*3.6)); // 135 km/h, dt=1s
    //this.LENGTH = (int)(2250/deltaX); // 2250 meter (300 sites with dx=7.5m)


    // the road as an array of speeds
    speed    = new int[LENGTH];
    // frequency of the gaps
    gapfreq  = new double[LENGTH];
    // frequency of the velocities
    vfreq    = new double[MAXSPEED+1];

    cars = 0;

    //calculate how fast the car closes the gap.
    setCloseGap();

    // clear road
    for (int i=0;i<LENGTH;i++) {
      speed[i] = -1;
    }

    vcolor = new Color[vfreq.length];
    // colors for different velocities
    for (int i= 0; i<vcolor.length; i++) {
      switch(i) {
      case 0: vcolor[i] = Color.red;    break;
      case 1: vcolor[i] = Color.blue;   break;
      case 2: vcolor[i] = Color.orange; break;
      case 3: vcolor[i] = Color.cyan;   break;
      case 4: vcolor[i] = Color.green;  break;
      case 5: vcolor[i] = Color.yellow; break;
      default:
        int grey = (200*i)/vfreq.length;
        vcolor[i] = new Color (grey, grey, grey);
      }
    }

    // put vehicles on the road
    int cars_abs = (int)(density*LENGTH);
    invcars = 1.0/((cars_abs > 0)? cars_abs: 1.0/DBL_INF);
    while (cars < cars_abs) {
      int i = (int)(Math.random()*LENGTH);
      if (speed[i] == -1) {
	speed[i] = 0;
	cars++;
      }
    }
  }

  public void setCloseGap() {
    double s, rh, U;

    ve = new int[LENGTH];
    for (int gap = 0; gap < LENGTH; gap++) {
      s = (gap+1) * deltaX;
      rh = 1000.0 / s;

      U = MAXSPEED/110.0 * (28.125*(1.0-rh/160.0) +
                            82.875*Math.exp (-Math.pow (rh/36.0,2.8)) -
                            1.0*Math.exp (-rh*28.125/160.0));

      ve[gap] = (int) (U+0.5);
      if (ve[gap] < 0) ve[gap] = 0;

    }
  }


  private int updateSpeed(int speed, int gap,
			  double prob_slowdown, double lambda) {
    int oldspeed = speed;

    speed += Math.floor(lambda*(ve[gap]-speed));

    // slowdown with probability prob_slowdown
    if ((speed > 0) && (Math.random() < prob_slowdown)) speed--;

    if (speed > gap) {
      // System.out.println ("crash! v: " + oldspeed + " gap: " + gap +
      //                     " -> v: " + speed);
      speed = gap;
    }
    return speed;
  }

  ////////////////////////////////////
  // update of the cellulare automaton

  public void update(double prob_slowdown, double prob_create,
		     double lambda, double changelane) {
    // number of vehicles
    int car_local = (int)(prob_create*LENGTH);
    // the different between the actual and the requested number of vehicles
    int diff_cars = cars-car_local;

    for(int i=0;i<LENGTH;i++) {
      gapfreq[i]=0.0;
    }
    for(int i=0;i<=MAXSPEED;i++) {
      vfreq[i]=0.0;
    }

    if (car_local > LENGTH) {
         car_local = LENGTH;
    }
    if (car_local < 1) {
       car_local = 1;
    }

    // match the number of vehicles
    if (diff_cars > 0) {
      // number of vehivles, that have to be killed
      int kill_cars = 0;
      while (kill_cars < diff_cars) {
	int i = (int)(Math.random()*LENGTH);
	if (speed[i] != -1) {
	  speed[i] = -1;
	  kill_cars++;
	}
      }
      cars = car_local;
      invcars  = 1.0/((cars > 0)? cars: 1.0/DBL_INF);
    } else if (diff_cars < 0) {
      // number of vehivles, that have to be created
      int new_cars = 0;
      while (new_cars < Math.abs(diff_cars))
	{
	  int i = (int)(Math.random()*LENGTH);
	  if (speed[i] == -1) {
	    int gap = i;
	    while (speed[++gap%LENGTH] == -1);
	    gap-=(i+1);
	    speed[i] = Math.min(gap,MAXSPEED);
	    new_cars++;
	  }
	}
      cars = car_local;
      invcars  = 1.0/((cars > 0)? cars: 1.0/DBL_INF);
    }

    // where is the 1st vehicle
    int i = -1;
    while((++i < LENGTH) && (speed[i] == -1));
    // go on until reaching the end of the lane
    while (i < LENGTH) {
      // searching for the vehicle ahead
      int gap = i;
      while (speed[++gap%LENGTH] == -1);

      // distance between two following vehicles
      gap-=(i+1);
      gapfreq[gap]+=((invcars < DBL_INF)? invcars: 0);

      ////////////////
      // update rules

      speed[i] = updateSpeed (speed[i], gap, prob_slowdown, lambda);
      vfreq[speed[i]]+=((invcars < DBL_INF)? invcars: 0);

      // next vehicle is on site j
      i+=(gap+1);
    }

    // move the vehicles
    // where is the 1st vehicle
    car_local = 0;
    i = -1;
    while((++i < LENGTH) && (speed[i] == -1));
    // go on until reaching the end of the lane
    while (i < LENGTH) {
      car_local++;
      int inext = i+speed[i];
      int j = inext%LENGTH;
      // exchange the positions
      if (i != j) {
	speed[j] = speed[i];
	speed[i] = -1;
      }
      // searching for the vehicle ahead
      while (speed[++inext%LENGTH] == -1);
      i = inext;
    }
  }

  public void paint(Graphics g, int row,
                                int dotdist, int dotsize, int xshift) {
      int i, j, num, local_speed;

      for (i = 0; i < 400; i++) {
        num = 0; local_speed = 0;
        //for (j = 0; j < combine; j++) {
          if (speed[i] >= 0) {
            local_speed += speed[i];
            num++;
         // }
        }
        if (num > 0) {
          g.setColor(vcolor[local_speed/num]);
          g.fillRect((xshift+i)*dotdist,row,dotsize,dotsize);
        }
      }
      g.setColor(background);
  }


  public int[] getPreviousRules()
    {
      int[] local_speed = new int[LENGTH];
      for(int i = 0;i<speed.length;i++){
        if(speed[i] > MAXSPEED)
          speed[i] = MAXSPEED;
        local_speed[i] = speed[i];
      }
      return local_speed;
   }

}  // End of "class Road"
